Module all_modules

! --------------------------------------------------------------------------
! MPI Version of the Spherical Acoustic Sun Simulator.
! Copyright 2006, Shravan Hanasoge
                                                                                                                                                        
! Hansen Experimental Physics Laboratory
! 455 Via Palou way, Stanford
! CA 94305, USA
! Email: shravan@stanford.edu
! --------------------------------------------------------------------------
!
! More Subroutines.
!

use initialize
implicit none

Contains
!==================================================================================

function norm2(matrix)
  
   implicit none
   integer i,j,ierr
   real*8 matrix(nx,dim2(rank))
   real*8 norm2, sum

   sum = 0.0  
   norm2  = 0.0
   do j =1,dim2(rank)
    do i =1,nx

      sum = sum + matrix(i,j)**2.0
    end do     
   end do     

   call MPI_REDUCE(sum, norm2, 1, MPI_DOUBLE_PRECISION, &
				MPI_SUM, 0, MPI_COMM_WORLD, ierr)
   

   norm2 = (norm2/(DBLE(nx)*DBLE(ny)))**0.5

end function norm2

!================================================================================

      SUBROUTINE printerror(status)

      ! See cookbook.f on FITSIO for details

      integer status
      character errtext*30,errmessage*80

      if (status .le. 0)return

      call ftgerr(status,errtext)
      print *,'FITSIO Error Status =',status,': ',errtext

      call ftgmsg(errmessage)
      do while (errmessage .ne. ' ')
          print *,errmessage
          call ftgmsg(errmessage)
      end do

      end SUBROUTINE printerror

!================================================================================
      SUBROUTINE deletefile(filename,status)

      integer status,unit,blocksize
      character*(*) filename

      if (status .gt. 0)return
      call ftgiou(unit,status)
      call ftopen(unit,filename,1,blocksize,status)
      

      if (status .eq. 0)then
          call ftdelt(unit,status)
      endif
      if (status .eq. 103)then
          status=0
          call ftcmsg
      endif 
      if ((status .NE. 0) .and. (status .ne. 103)) then 

          status=0
          call ftcmsg
          call ftdelt(unit,status)
      end if

      call ftfiou(unit, status)

      end SUBROUTINE deletefile

!================================================================================

   SUBROUTINE readfits(filename,read,dim3)

      implicit none
      integer status,unit,readwrite,blocksize,naxes(3), tag, k, dim3 
      integer group,firstpix, ierr, stat(MPI_STATUS_SIZE)
      integer disp, temptyp, sendtyp
      integer nelements
      real*8 nullval,read(nx,dim2(rank),dim3)
      real*8, dimension(:,:,:), allocatable :: temp
      logical anynull
      character*(*) filename
      integer (KIND=MPI_ADDRESS_KIND) ext, dum, spacing

      
      if (rank == 0) then
       allocate(temp(nx,ny,dim3))
       status=0
       call ftgiou(unit,status)
       readwrite=0
       print *,'Now reading the file: '//filename
       call ftopen(unit,filename,readwrite,blocksize,status)

       naxes(1) = nx
       naxes(2) = ny
       naxes(3) = dim3
       nelements=naxes(1)*naxes(2)*naxes(3)
       group=1
       firstpix=1
       nullval=-999

       call ftgpvd(unit,group,firstpix,nelements,nullval, &
       &            temp,anynull,status)


       call ftclos(unit, status)
       call ftfiou(unit, status)

       if (status .gt. 0)call printerror(status)

       read = temp(:,1:dim2(0),:)  

       call MPI_TYPE_GET_EXTENT(MPI_DOUBLE_PRECISION, dum, ext, ierr)
       spacing = nx*ny*ext
       disp = dim2(0) + 1
      endif

      call MPI_BARRIER(MPI_COMM_WORLD, ierr)       
      do k=1,numtasks-1
                     
        if (rank ==0) then
 	 call MPI_TYPE_VECTOR(nx*dim2(k),1,1,MPI_DOUBLE_PRECISION,temptyp,ierr)
	 call MPI_TYPE_CREATE_HVECTOR(dim3,1,spacing,temptyp,sendtyp,ierr)
	 call MPI_TYPE_COMMIT(sendtyp, ierr)
	 tag = 0
	 call MPI_SEND(temp(1,disp,1), 1, sendtyp, &
			k, tag, MPI_COMM_WORLD, ierr)
	 disp = disp + dim2(k)
        endif
        if (rank  == k) then
  	 nelements = nx * dim2(k) * dim3
	 tag = 0
	 call MPI_RECV(read, nelements, MPI_DOUBLE_PRECISION,&
			0, tag, MPI_COMM_WORLD, stat, ierr)
        endif

        call MPI_BARRIER(MPI_COMM_WORLD, ierr)
 
      enddo

      call MPI_BARRIER(MPI_COMM_WORLD, ierr)
      if (rank == 0) deallocate(temp)

    end SUBROUTINE readfits

!================================================================================


function norm(matrix)

   implicit none
   integer i, j, k, ierr
   real*8 matrix(nx,dim2(rank),nz)
   real*8 norm, sum
  
   norm  = 0.0
   sum = 0.0

   do k = 1,nz
    do j =1,dim2(rank)
     do i =1,nx
       sum = sum + matrix(i,j,k)**2.0
     end do     
    end do     
   end do 
 
   call MPI_REDUCE(sum, norm, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 0, MPI_COMM_WORLD, ierr) 

   norm = (norm/(DBLE(nx)*DBLE(ny)*DBLE(nz)))**0.5d0

   call MPI_BARRIER(MPI_COMM_WORLD, ierr)
end function norm

!================================================================================

 SUBROUTINE convert_to_string(numbe,sting,length_string)

  implicit none
  integer i,length_string,numbe,n(1:length_string),number_temp
  character*(length_string) sting
  character*1 charc(10)

  charc(1)  = '0'
  charc(2)  = '1'
  charc(3)  = '2'
  charc(4)  = '3'
  charc(5)  = '4'
  charc(6)  = '5'
  charc(7)  = '6'
  charc(8)  = '7'
  charc(9)  = '8'
  charc(10) = '9'


  number_temp = numbe
  do i=length_string,1,-1

    n(length_string-i+1) = floor(number_temp*10.0**(-(i-1.0)))
    number_temp = number_temp - n(length_string-i+1)*10**(i-1)
    sting(length_string-i+1:length_string-i+1) = charc(n(length_string-i+1)+1)

  enddo

  end SUBROUTINE convert_to_string

!================================================================================


function norm_1D(matrix)

   use initialize   
   implicit none
   integer i, j, k, ierr
   real*8 matrix(nz)
   real*8 norm_1d, sum

  
   norm_1d  = 0.0
   sum = 0.0
   do k = 1,nz
     sum = sum + matrix(k)**2.0
   end do 

   norm_1d = (sum/DBLE(nz))**0.5


end function norm_1d

!================================================================================

     function date_time()
  
      implicit none 
      character*27 date_time
      integer i, j, one_day, day, month, hour, minutes
      parameter(one_day = 1440)
      character*2 monthchar, daychar, hourchar, minutechar, secondchar
      character*4 yearchar

      ! Assuming the timestep (in seconds) divides 60 seconds.

      yearchar = '2007'
      monthchar = '06'
      secondchar = '00'
      day = time/(steps*one_day) ! Number of days of computation
      hour = (time/steps - day*1440)/60 ! Remainder hours
      minutes =  (time/steps - day*1440 - hour*60) ! Remainder number of minutes
      day = day + 1 ! Starting day is first day of the month

      call convert_to_string(day, daychar, 2)
      call convert_to_string(hour, hourchar, 2)
      call convert_to_string(minutes, minutechar, 2)

!      date_time ='2015.05.11_'//t1//':'//t2//':'01':00.0_UT'
    
      date_time = yearchar//'.'//monthchar//'.'//daychar//'_'//hourchar//':'//minutechar//':'//secondchar//':00.0_UT'
     end function date_time

!================================================================================
     SUBROUTINE writefits_local(filename, dump_array, dim3)

      implicit none
      integer blocksize,bitpix,naxes(3),unit1,dim3
      integer status1,group,fpixel, ierr
      integer temptype, sendtyp
      integer*8 nelements
      real*8 dump_array(nx,dim2(rank),dim3)
      character*(*) filename
      logical simple,extend

      status1 = 0
      call ftgiou(unit1,status1)
      blocksize=1
      call ftinit(unit1,filename,blocksize,status1)
      simple=.true.
      bitpix=-64
      naxes(1)=nx
      naxes(2)=dim2(rank)
      naxes(3)=dim3
      nelements=naxes(1)*naxes(2)*naxes(3)
      extend=.false.
      group=1
      fpixel=1
                                                                                                                                                      
      call ftphpr(unit1,simple,bitpix,3,naxes,0,1,extend,status1)
      call ftpprd(unit1,group,fpixel,nelements,dump_array,status1)
      call ftclos(unit1, status1)
      call ftfiou(unit1, status1)

      if (status1 .gt. 0) call printerror(status1)

!      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

     end SUBROUTINE writefits_local
!===============================================================================

     SUBROUTINE writefits_3D(filename,dump_array,dim3)

      implicit none 
      integer blocksize,bitpix,naxes(3),unit1,k,disp, tag, dim3
      integer status1,group,fpixel,ierr,stat(MPI_STATUS_SIZE)
      integer temptype, sendtyp, nelements
      real*8 dump_array(nx,dim2(rank),dim3)
      real*8, allocatable, dimension(:,:,:) :: temp
      character*(*) filename
      logical simple,extend
      integer (KIND = MPI_ADDRESS_KIND) space, dum, ext
 
      if (rank == 0) then
	 allocate(temp(nx,ny,dim3))
         temp(:, 1:dim2(0), :) = dump_array
         disp = dim2(0) + 1
         print *,'Writing file', filename
      endif
      call MPI_BARRIER(MPI_COMM_WORLD,ierr)

      do k=1,numtasks-1

       if (rank == k) then
	 tag = 0
         nelements = nx*dim3*dim2(rank)
         call MPI_SEND(dump_array, nelements, MPI_DOUBLE_PRECISION, &
			0, tag, MPI_COMM_WORLD, ierr)
       endif
       if (rank == 0) then

	  call MPI_TYPE_GET_EXTENT(MPI_DOUBLE_PRECISION,dum, ext,ierr)
	  space = nx*ny*ext
	  tag = 0
	  call MPI_TYPE_VECTOR(nx*dim2(k),1,1,MPI_DOUBLE_PRECISION,temptype,ierr)
	  call MPI_TYPE_CREATE_HVECTOR(dim3,1,space,temptype,sendtyp,ierr)
	  call MPI_TYPE_COMMIT(sendtyp, ierr)
	  call MPI_RECV(temp(1,disp,1), 1, sendtyp, &
			k, tag, MPI_COMM_WORLD, stat, ierr)

	  disp = disp + dim2(k)
        endif
 	call MPI_BARRIER(MPI_COMM_WORLD,ierr)
       enddo
      
       if (rank == 0) then
         status1 = 0
         call ftgiou(unit1,status1)
         blocksize=1

         call ftinit(unit1,filename,blocksize,status1)
         simple=.true.
         bitpix=-64
         naxes(1)=nx
         naxes(2)=ny
         naxes(3)=dim3
         nelements=naxes(1)*naxes(2)*naxes(3)
         extend=.false.
         group=1
         fpixel=1
         
         call ftphpr(unit1,simple,bitpix,3,naxes,0,1,extend,status1)
         call ftpkyj(unit1,'TIME',time,'Also, init = time+1', status1)
         call ftpprd(unit1,group,fpixel,nelements,temp,status1)
         call ftclos(unit1, status1)
         call ftfiou(unit1, status1)

         deallocate(temp)
         if (status1 .gt. 0) call printerror(status1)

       endif

       call MPI_BARRIER(MPI_COMM_WORLD,ierr)

     end SUBROUTINE writefits_3D
!================================================================================
function sum2(matrix, dima, dimb, option)
  
   implicit none
   integer i,j,ierr, option, dima, dimb
   real*8 matrix(dima, dimb)
   real*8 sum2, summer

   summer = 0.0  
   sum2  = 0.0
   if (option == 0) then
    do j =1,dimb
     do i =1,dima
       summer = summer + matrix(i,j)
     end do     
    end do     
   else
    do j =1,dimb
     do i =1,dima
       summer = summer + abs(matrix(i,j))
     end do     
    end do     
   endif

   call MPI_REDUCE(summer, sum2, 1, MPI_DOUBLE_PRECISION, &
				MPI_SUM, 0, MPI_COMM_WORLD, ierr)
   

   sum2 = (sum2/(DBLE(dima)*DBLE(dimb)))

end function sum2

!================================================================================
function sum2_normal(matrix, dima, dimb, option)
  
   implicit none
   integer i,j,ierr, option, dima, dimb
   real*8 matrix(dima, dimb)
   real*8 sum2_normal, summer

   summer = 0.0  
   sum2_normal  = 0.0
   if (option == 0) then
    do j =1,dimb
     do i =1,dima
       summer = summer + matrix(i,j)
     end do     
    end do     
   else
    do j =1,dimb
     do i =1,dima
       summer = summer + abs(matrix(i,j))
     end do     
    end do     
   endif

   sum2_normal = (summer/(DBLE(dima)*DBLE(dimb)))

end function sum2_normal


!================================================================================

   SUBROUTINE readfits_1proc(filename,read,dim1,dim2,dim3)

      implicit none
      integer status,unit,readwrite,blocksize,naxes(3), tag, k, dim3 ,dim2,dim1
      integer group,firstpix, ierr, stat(MPI_STATUS_SIZE)
      integer disp, temptyp, sendtyp
      integer nelements
      real*8 nullval,read(dim1,dim2,dim3)
      real*8, dimension(:,:,:), allocatable :: temp
      logical anynull
      character*(*) filename
      integer (KIND=MPI_ADDRESS_KIND) ext, dum, spacing

      
      status=0
       call ftgiou(unit,status)
       readwrite=0
       print *,'Now reading the file: '//filename
       call ftopen(unit,filename,readwrite,blocksize,status)

       naxes(1) = dim1
       naxes(2) = dim2
       naxes(3) = dim3
       nelements=naxes(1)*naxes(2)*naxes(3)
       group=1
       firstpix=1
       nullval=-999

       call ftgpvd(unit,group,firstpix,nelements,nullval, &
       &            read,anynull,status)


       call ftclos(unit, status)
       call ftfiou(unit, status)

       if (status .gt. 0)call printerror(status)



    end SUBROUTINE readfits_1proc

!================================================================================
  
  SUBROUTINE READ_IN_FIELD ()

   implicit none
   integer j

   call readfits(dirbackmag//'density.fits',rho0,nz)
   call readfits(dirbackmag//'pressure.fits',p0,nz)
   call readfits(dirbackmag//'B_x.fits',box,nz)
   call readfits(dirbackmag//'B_y.fits',boy,nz)
   call readfits(dirbackmag//'B_z.fits',boz,nz)

   boz =  boz/dimb
   box = box/dimb
   boy = boy/dimb

   p0 = p0/(dimrho * dimc**2.0)
   rho0 = rho0/dimrho

   do j=1,nz
    c2(:,:,j) = gamma(j)*p0(:,:,j)/rho0(:,:,j)
   enddo
   c_speed= c2**(0.5)
  
  END SUBROUTINE READ_IN_FIELD
   
!================================================================================
  
  SUBROUTINE READ_IN_FLOWS ()
  
   implicit none

   call readfits(dirbackmag//'v0_x.fits',v0_x,nz)
   call readfits(dirbackmag//'v0_y.fits',v0_y,nz)
   call readfits(dirbackmag//'v0_z.fits',v0_z,nz)

   v0_x = v0_x/dimc
   v0_y = v0_y/dimc
   v0_z = v0_z/dimc

  END SUBROUTINE READ_IN_FLOWS

!================================================================================

  SUBROUTINE READ_IN_INITIAL_CONDITION ()
 
    implicit none
    character*(timestamp_size) tempc
  
    call convert_to_string(time0,tempc,timestamp_size)

    if (.not. displ .and. (.not. TEST_IN_2D)) then

     call readfits(directory//'p_'//tempc//'_full.fits',a(:,:,:,5),nz)
     call readfits(directory//'vy_'//tempc//'_full.fits',a(:,:,:,3),nz)
     call readfits(directory//'rho_'//tempc//'_full.fits',a(:,:,:,1),nz)
     call readfits(directory//'vx_'//tempc//'_full.fits',a(:,:,:,2),nz)
     call readfits(directory//'vz_'//tempc//'_full.fits',a(:,:,:,4),nz)

     a(:,:,:,5) = a(:,:,:,5)/(dimc**2. * dimrho)
     a(:,:,:,1) = a(:,:,:,1)/dimrho
     a(:,:,:,2) = a(:,:,:,2)/dimc
     a(:,:,:,3) = a(:,:,:,3)/dimc
     a(:,:,:,4) = a(:,:,:,4)/dimc

     if (magnetic) then
      call readfits(directory//'bx_'//tempc//'_full.fits',a(:,:,:,6),nz)
      call readfits(directory//'by_'//tempc//'_full.fits',a(:,:,:,7),nz)
      call readfits(directory//'bz_'//tempc//'_full.fits',a(:,:,:,8),nz)

      a(:,:,:,6) = a(:,:,:,6)/dimb
      a(:,:,:,7) = a(:,:,:,7)/dimb
      a(:,:,:,8) = a(:,:,:,8)/dimb

     endif

    elseif (displ) then

     call readfits(directory//'xix_'//tempc//'_full.fits',a(:,:,:,1),nz)
     call readfits(directory//'xiy_'//tempc//'_full.fits',a(:,:,:,2),nz)
     call readfits(directory//'xiz_'//tempc//'_full.fits',a(:,:,:,3),nz)
     call readfits(directory//'vx_'//tempc//'_full.fits',a(:,:,:,4),nz)
     call readfits(directory//'vy_'//tempc//'_full.fits',a(:,:,:,5),nz)
     call readfits(directory//'vz_'//tempc//'_full.fits',a(:,:,:,6),nz)

     a(:,:,:,1) = a(:,:,:,1)/diml
     a(:,:,:,2) = a(:,:,:,2)/diml
     a(:,:,:,3) = a(:,:,:,3)/diml
     a(:,:,:,4) = a(:,:,:,4)/dimc
     a(:,:,:,5) = a(:,:,:,5)/dimc
     a(:,:,:,6) = a(:,:,:,6)/dimc

    endif

    if (USE_PML) then
     call readfits(directory//'psivz_'//tempc//'_full.fits',pmlvars(:,:,:,1),nzpml)
     call readfits(directory//'psip_'//tempc//'_full.fits',pmlvars(:,:,:,2),nzpml)
     if (magnetic) then         
      call readfits(directory//'psidzbx_'//tempc//'_full.fits',pmlvars(:,:,:,3),nzpml)
      call readfits(directory//'psidzby_'//tempc//'_full.fits',pmlvars(:,:,:,4),nzpml)
      call readfits(directory//'psiinductionbx_'//tempc//'_full.fits',pmlvars(:,:,:,5),nzpml)
      call readfits(directory//'psiinductionby_'//tempc//'_full.fits',pmlvars(:,:,:,6),nzpml)
     endif
    endif

  END SUBROUTINE READ_IN_INITIAL_CONDITION


!================================================================================

  SUBROUTINE WRITE_OUT_FULL_STATE

    implicit none
    character*(timestamp_size) tempc
  
    call convert_to_string(time,tempc,timestamp_size)

    if (.not. displ) then

     call writefits_3d(directory//'p_'//tempc//'_full.fits',a(:,:,:,5)*dimrho*dimc**2.,nz)
     call writefits_3d(directory//'vy_'//tempc//'_full.fits',a(:,:,:,3)*dimc,nz)
     call writefits_3d(directory//'rho_'//tempc//'_full.fits',a(:,:,:,1)*dimrho,nz)
     call writefits_3d(directory//'vx_'//tempc//'_full.fits',a(:,:,:,2)*dimc,nz)
     call writefits_3d(directory//'vz_'//tempc//'_full.fits',a(:,:,:,4)*dimc,nz)

     if (magnetic) then
      call writefits_3d(directory//'bx_'//tempc//'_full.fits',a(:,:,:,6)*dimb,nz)
      call writefits_3d(directory//'by_'//tempc//'_full.fits',a(:,:,:,7)*dimb,nz)
      call writefits_3d(directory//'bz_'//tempc//'_full.fits',a(:,:,:,8)*dimb,nz)
     endif

    elseif (displ) then

     call writefits_3d(directory//'xix_'//tempc//'_full.fits',a(:,:,:,1)*diml,nz)
     call writefits_3d(directory//'xiy_'//tempc//'_full.fits',a(:,:,:,2)*diml,nz)
     call writefits_3d(directory//'xiz_'//tempc//'_full.fits',a(:,:,:,3)*diml,nz)
     call writefits_3d(directory//'vx_'//tempc//'_full.fits',a(:,:,:,4)*dimc,nz)
     call writefits_3d(directory//'vy_'//tempc//'_full.fits',a(:,:,:,5)*dimc,nz)
     call writefits_3d(directory//'vz_'//tempc//'_full.fits',a(:,:,:,6)*dimc,nz)

    endif

    if (USE_PML .and. (.not. TEST_IN_2D)) then
     call writefits_3d(directory//'psivz_'//tempc//'_full.fits',pmlvars(:,:,:,1),nzpml)
     call writefits_3d(directory//'psip_'//tempc//'_full.fits',pmlvars(:,:,:,2),nzpml)
     if (magnetic) then         
      call writefits_3d(directory//'psidzbx_'//tempc//'_full.fits',pmlvars(:,:,:,3),nzpml)
      call writefits_3d(directory//'psidzby_'//tempc//'_full.fits',pmlvars(:,:,:,4),nzpml)
      call writefits_3d(directory//'psiinductionbx_'//tempc//'_full.fits',pmlvars(:,:,:,5),nzpml)
      call writefits_3d(directory//'psiinductionby_'//tempc//'_full.fits',pmlvars(:,:,:,6),nzpml)
     endif
    endif


  END SUBROUTINE WRITE_OUT_FULL_STATE


!================================================================================

  SUBROUTINE WRITE_OUT_SLICE(radial)

    implicit none
    integer radial
    character*(timestamp_size) tempc
  
    call convert_to_string(time,tempc,timestamp_size)

    if (.not. displ .and. (.not. TEST_IN_2D)) then

     call writefits_3d(directory//'p_'//tempc//'_slice.fits',a(:,:,radial,5)*dimrho*dimc**2.,1)
     call writefits_3d(directory//'vy_'//tempc//'_slice.fits',a(:,:,radial,3)*dimc,1)
     call writefits_3d(directory//'rho_'//tempc//'_slice.fits',a(:,:,radial,1)*dimrho,1)
     call writefits_3d(directory//'vx_'//tempc//'_slice.fits',a(:,:,radial,2)*dimc,1)
     call writefits_3d(directory//'vz_'//tempc//'_slice.fits',a(:,:,radial,4)*dimc,1)

     if (magnetic) then
      call writefits_3d(directory//'bx_'//tempc//'_slice.fits',a(:,:,radial,6)*dimb,1)
      call writefits_3d(directory//'by_'//tempc//'_slice.fits',a(:,:,radial,7)*dimb,1)
      call writefits_3d(directory//'bz_'//tempc//'_slice.fits',a(:,:,radial,8)*dimb,1)
     endif

    elseif (displ .and. (.not. TEST_IN_2D)) then

     call writefits_3d(directory//'xix_'//tempc//'_slice.fits',a(:,:,radial,1)*diml,1)
     call writefits_3d(directory//'xiy_'//tempc//'_slice.fits',a(:,:,radial,2)*diml,1)
     call writefits_3d(directory//'xiz_'//tempc//'_slice.fits',a(:,:,radial,3)*diml,1)
     call writefits_3d(directory//'vx_'//tempc//'_slice.fits',a(:,:,radial,4)*dimc,1)
     call writefits_3d(directory//'vy_'//tempc//'_slice.fits',a(:,:,radial,5)*dimc,1)
     call writefits_3d(directory//'vz_'//tempc//'_slice.fits',a(:,:,radial,6)*dimc,1)

    endif


  END SUBROUTINE WRITE_OUT_SLICE


!================================================================================

end module all_modules

