Program driver

 ! --------------------------------------------------------------------------
 ! MPI Version of the Cartesian Acoustic Sun Simulator. 
 ! Copyright 2006, Shravan Hanasoge
 
 ! Hansen Experimental Physics Laboratory
 ! 455 Via Palou way, Stanford
 ! CA 94305, USA
 ! Email: shravan@stanford.edu 

 ! The driver routine. It utilizes the other subroutines to perform time-
 ! advancements. The time advancement is performed with the explicit Low
 ! Dissipation, Low Dispersion Runge Kutta 4th order method, described in
 ! Hu et al. (1996). Radial derivatives are computed using compact finite
 ! difference (Lele 1992). Horizontal variations are extracted by projecting
 ! the desired quantity onto Fourier harmonic space followed by an
 ! appropriate inverse fourier transform.
 

 !
 ! BASIC CHECKS:
 ! 1. If init is all right.
 ! 2. If the time-length of the simulation matches
 ! 3. If the directories are OK
 ! 4. If the forcing function is being read from the right directory
 ! 5. If the variables are initialized OK
 ! 6. If the flow profiles are allright   
 ! 7. If output and status are deleted or appropriately moved
 !
 
 ! ---------------------------------------------------------------------------
  use initialize
  use all_modules
  use mp_physics
  use time_step
  use damping

  implicit none

    integer i,j, init, ierr, t, k, index(1000,2), randinit(2), aind
    real*8 start_time,end_time,tempf,t1,T00,nor1,nor2,nor3,nor4,nor5,e, mp_time
    real*8 data(nz,6),total_time, start_mp_time, avg_time,zz, tempxy, rand1, rand2
    logical saved, first_iteration, init_variables 
    character*5 tempc
    real*8, allocatable, dimension(:,:,:) :: tempo,check,outpu    
 20 format(8f12.4)


 ! Initializing the computation
 ! --------------------------------------------

 call Initialize_all()

 start_mp_time = MPI_WTIME()
 start_time = MPI_WTIME()

 !!!!!!! SOUND-SPEED PERTURBATIONS HERE !!!!!!

 if (.not. TEST_IN_2D) call mp_initialize_RHS()
 if (TEST_IN_2D) call mp_initialize_RHS_2d ()

 call Initialize_step()

 if (DAMP_WAVES) call INIT_DAMPING 

 end_time = MPI_WTIME()

 if (rank == 0) &
  print *, 'Initialization took: ', (end_time -start_time)

 !--------------------------------------------- 

 if (time0 .ne.0) call read_in_initial_condition

!!!! OTHER INITIAL CONDITIONS!!!!!!
! a(:,:,:,1) - density
! a(:,:,:,2) - v_x
! and so on - v_y, v_z, p, b_x, b_y, b_z

 ! Length of simulation

  maxtime = floor(solartime*3600.0/timestep) + 1
  maxtime = 2*(maxtime/2) 
  total_time = wall_time * 3600.

  steps = floor(outputcad/timestep + 0.1)
  freq_filtering = floor(60./timestep + 0.1)

  if (rank ==0) then
    print *,'FREQUENCY OF OUTPUT AT', steps, 'TIMESTEPS'
    print *, 'NUMBER OF TIMESTEPS =', maxtime
    print *, 'TIMESTEP =', timestep 
  endif

  ! Excitation series computation -----------------------------

  num_steps = FLOOR(DBLE(maxtime)*timestep/cadforcing) + 2
  num_steps = 2*(num_steps/2)+2
  saved = .false.

  if (.not. TEST_IN_2D) then
   allocate(vr(nx,dim2(rank),num_steps))

   call readfits(forcingfunc,vr,num_steps)

   ! SOURCE APODIZATION (TEMPORAL AND SPATIAL)
   ! Note if periodic horizontal boundaries are not, there will be no
   ! horizontal spatial apodization
    do k=1,num_steps
     do j=1,dim2(rank)
      do i=1,nx
       tempxy=(((x(i)-0.5)*200)**2.0+((y(j+ystart(rank)-1)-0.5)*200)**2.0)**0.5
       vr(i,j,k) = vr(i,j,k) * (1.-spongex(i)) * (1. - spongey(j)) /(1.0+ exp((5.-k))) !* &
	     !1.0/(1.0+exp((10.0-tempxy)*1.5))
  
      enddo
     enddo
    enddo
   endif
  
   time_old = 0

  !!!!! SOURCE STRENGTH PERTURBATIONS HERE (i.e., in the above loop)!!!!!! 



   !----------------Driver Kernel----------------------------
   ! RMS_HIST STORES THE RMS TIME EVOLUTION OF v_z
   ! A VERY USEFUL INDICATOR OF STABILITY

  if (rank == 0) &
	open(19,file='rms_hist',status='replace',form='formatted',position='append') 

  if (time0 == 0) init = 0
  if (time0 .NE. 0) init = time0+1

  do i = init,maxtime

    time = i

  ! code to be timed
    call cpu_time(t1)
    T00 = t1
    start_time = MPI_WTIME() 
    call lagrange_interp ()
    call step()
    end_time = MPI_WTIME() 
    call cpu_time(t1) 
   
    nor1 = norm(a(:,:,:,4)) !* dimc
 
    mp_time = MPI_WTIME() 

  ! Timing and 'convergence' statistics 
  ! The criterion used to see if the results are 'convergent' is the
  ! L2 norm of the radial velocity. Quite silly, really.

   if (rank ==0) then
 
    print *, '----------------------------------------------------------'
    print *, '	Iteration #,  Cpu Time and Real Time:'
    print *,       time,       	(t1 - T00),        (time+1)*timestep
    print *
    print *,'	Wall clock time', (end_time - start_time)
    print *
    print *,nor1
    print *, '----------------------------------------------------------'
 

    write(19,*) nor1
!    if (nor1 .GT. 10.0**(-3.0)) stop  

   endif

   if (mod(time,steps) == 0) then

       if (rank == 0) print *, 'Writing output'

  ! LOOK AT ALL_MODULES 
  ! o_Rad is the radial grid point associated with observation height

       if (.not. test_in_2D) call write_out_slice(o_Rad)
       if (test_in_2D) call write_out_full_state

   endif


   if (( ((mp_time - start_mp_time) >  (total_time - 150.0)) .OR. (mod(time+1,500) == 0) )) then

     call write_out_full_state

     if (rank ==0 ) print *,'Full output done, index:',time
     if ((mp_time - start_mp_time) >  (total_time - 150.0)) stop

   endif

 end do

 call MPI_BARRIER(MPI_COMM_WORLD, ierr)

 call dfftw_destroy_plan(fftw_plan_fwd_x)
 call dfftw_destroy_plan(fftw_plan_inv_x)

 call dfftw_destroy_plan(fftw_plan_fwd_y)
 call dfftw_destroy_plan(fftw_plan_inv_y)

 if (rank == 0) then
   close(19)
 endif

 call MPI_BARRIER(MPI_COMM_WORLD, ierr) 
 call MPI_FINALIZE(ierr)


end program driver
