Module mp_physics_2d

! --------------------------------------------------------------------------
! MPI Version of the Cartesian Solar Wave Simulator.
! Copyright 2008, Shravan Hanasoge
                                                                                                                                                         
! W. W. Hansen Experimental Physics Laboratory
! Stanford University, Stanford, CA 94305, USA
! Email: shravan@solar.stanford.edu
! --------------------------------------------------------------------------

  use initialize
  use all_modules
  use mp_physics
  use derivatives
  implicit none

Contains


!============================================================================

 SUBROUTINE mp_Initialize_RHS_2d()

   implicit none
   integer i,j,k,l,flag(2),trans
   real*8 temp_r,temp_t,temp_p,in(nz), out(nz),advect0(nx,ny,nz),alphaz(nz), cmax
   real*8 temp,output,temp2,start_damping,cchar, finish_damping, tempxy
   real*8, allocatable, dimension(:,:,:) :: tempin, tempout, dampz, functiondecay

   ! Determination of excitation location.

   temp2 = (1.0 + excitdep/Rsun)
   temp2 = 0.9997
   temp = 1.0
   do k=1,nz
    if (abs(z(k) - temp2) .LE. temp) then
     temp = abs(z(k) - temp2)
     e_rad = k
    endif
   enddo

   ! Determination of the output height 
   temp_r = 1.0 + obsheight/rsun
   temp = 1.0
   do k=1,nz
    if (abs(z(k) - temp_r) .LE. temp) then
     temp = abs(z(k) - temp_r)
     o_rad = k
    endif
   enddo

   ! Computing the absorbent boundary adjoining sponge.
 


   do k=1,nz
    ! Sponge in the z-direction
    spongez(k) =  200.0/(1.0 + exp((1.002 - z(k))/0.0001)) & 
		+ 300.0/(1.0 + exp((z(k) - (z(1) + 0.005))/0.0005))
! Source depth is where the sources are located
    source_dep(k) = exp(-10.0**(7.0)*4.0 *(z(k) - z(e_rad))**2.0)

   enddo


   ! ALLOCATE FORCING FUNCTION STUFF

   maxtime = floor(solartime*3600.0/timestep) + 1
   maxtime = 2*(maxtime/2) 


   num_steps = FLOOR(DBLE(maxtime)*timestep/cadforcing) + 2
   num_steps = 2*(num_steps/2)+2

   allocate(tempin(nx, nx, num_steps))
   allocate(vr(nx,dim2(rank),num_steps))

   call readfits_1proc(forcingfunc,tempin,nx,nx,num_steps)

   vr(:,1,:) = tempin(:,20,:)
 
   j=1
   do k=1,num_steps
     do i=1,nx
      tempxy=(((x(i)-0.5)*xlength*10.0**(-8.0))**2.0)**0.5
      vr(i,j,k) = vr(i,j,k) /(1.0+ exp((5.-k))) * &
	     1.0/(1.0+exp((10.0-tempxy)*1.5))
 
     enddo
   enddo

   deallocate(tempin)
  !!!!!!!!!!!!!!!!

   if (magnetic) then
    call create_2d_slices ()

    do k=1,nz
     reduction(:,:,k) = (boz(:,:,k)**2.+ box(:,:,k)**2.)/(rho0(:,:,k)*cq(k)**2.) 
    enddo
    
    reduction = 8./(8. + reduction)

    do k=nz-10, nz
     reduction(:,:,k) = reduction(:,:,k) * ((k - nz)/10.0)**4.
    enddo
  
    box = box * reduction**0.5
    boz = boz * reduction**0.5
    
    call ddx(boz, reduction,1)
    call ddz(box, curlboy, 1)
    do k=nz-30,nz 
     curlboy(:,:,k) = curlboy(:,:,k)/(1. + exp(k-(nz-20.0)) )
    enddo
    curlboy = - reduction
    
    deallocate(reduction)

   endif
    
 
   ! Damping length from the horizontal boundaries
   start_damping = 10.0*10.0**(8.0)/xlength
   finish_damping = 1. -  10.0*10.0**(8.0)/xlength

  spongex = 0.0
  spongey = 0.0

  if (.NOT. PERIODIC) then
   ! Sponge in the x-direction
    do i=1,nx
     spongex(i) = (1.0 - 1.0/( (1.0 + exp((start_damping-x(i))/0.008))*&
				(1.0 + exp((x(i)-finish_damping)/0.008)) ))
    enddo 

   ! Sponge in the distributed y-direction
   endif

! Total sponge
   spongex = 0.0
   spongey = 0.0
   if (PERIODIC) then 
     do k=1,nz
     do j=1,dim2(rank)
      spongexyz(:,j,k) = spongez(k)
     enddo
    enddo
   else
     do k=1,nz
     do j=1,dim2(rank)
      spongexyz(:,j,k) = (spongez(k) + 100.*spongex(:))
     enddo
    enddo
   endif 

   rhoinv = 1.0/rho0
   c2rho0 = c2 * rho0
   
   call ddz(rho0, gradrho0_z, 1)
   call ddz(p0, gradp0_z, 1)


! Stabilize the background model

   do k=1,nz
      do j=1,dim2(rank)
         do i=1,nx

            ! FOR NEAR SURFACE LAYERS ONLY, (-0.15<z<0.1)                                                     
            gradp0_z(i,j,k) =max(c2(i,j,k)*gradrho0_z(i,j,k)+10.**(-5.)*Rsun/(dimrho*dimc**2.),-rho0(i,j,k)*g(k))

            ! AND FOR THE REST OF THE BOX z<= -0.15 and z>=0.1 Mm
            if (z(k) .LE. 0.9998) then
               gradp0_z(i,j,k) =max(c2(i,j,k)*gradrho0_z(i,j,k),-rho0(i,j,k)*g(k))
            endif

            if (z(k) .GE. 1.0001) then
               gradp0_z(i,j,k) =max(c2(i,j,k)*gradrho0_z(i,j,k),-rho0(i,j,k)*g(k))
            endif

         enddo
      enddo
   enddo

   if (use_pml) then
    allocate(functiondecay(nx,dim2(rank),nz),dampz(nx,dim2(rank),nz),tempin(nx,dim2(rank),nz))
    spongexyz = 0.
 
    if (magnetic) then
     do k=1,nz
      tempin(:,:,k) = (c2(:,:,k) + (box(:,:,k)**2.+ boz(:,:,k)**2.)/rho0(:,:,k))**0.5
     enddo
    endif
    if (.not. magnetic) tempin = c2**0.5

    dampz = 0.
    alphaz = pi*0.005*diml/dimc
    functiondecay = 1.

    do k=1,nz
     if (k < npmlbot+1)  then
      alphaz(k) = pi * 0.005 * diml/dimc*(z(k)-z(1))/(z(npmlbot)-z(1))
      dampz(:,:,k) = 3.*3.*tempin(:,:,1)/(z(npmlbot) - z(1))*((z(npmlbot)-z(k))/(z(npmlbot)-z(1)))**2. !tempin(:,:,1)
      functiondecay(:,:,k) = alphaz(k)/(dampz(:,:,k) + alphaz(k)) 
      gradp0_z(:,:,k)  =gradp0_z(:,:,k) * functiondecay(:,:,k)
      gradrho0_z(:,:,k)  =gradrho0_z(:,:,k) * functiondecay(:,:,k)
      g(k)  =g(k) * functiondecay(1,1,k)
     endif

     if (k .gt. nz-npmltop) then
      alphaz(k) = pi * 0.005 * diml/dimc * (z(nz) -z(k))/(z(nz)-z(nz-npmltop+1))
      dampz(:,:,k) = 3.*3.*tempin(:,:,nz-5)/(z(nz) - z(nz-npmltop+1))*((z(nz-npmltop +1)-z(k))/(z(nz)-z(nz-npmltop+1)))**2. ! tempin(:,:,nz)
 
      functiondecay(:,:,k) = alphaz(k)/(dampz(:,:,k) + alphaz(k)) 
      gradp0_z(:,:,k)  =gradp0_z(:,:,k) * functiondecay(:,:,k)
      gradrho0_z(:,:,k)  =gradrho0_z(:,:,k) * functiondecay(:,:,k)
      g(k)  =g(k) * functiondecay(1,1,k)
     endif

     bzpml(:,:,k) = - alphaz(k) - dampz(:,:,k)
     az(:,:,k) = -dampz(:,:,k)


    enddo
    c_speed = c2**0.5

    deallocate(tempin, dampz, functiondecay)

    scrpml = 0.0 
    pmlvars = 0.0
    psipml = 0.0
   endif

   if (magnetic) then
 
    call ddx(rho0, gradrho0_x ,1)
    call ddx(p0, gradp0_x ,1)

   endif
  
   if (rank == 0) then 
     print *,'Source excitation at radius', z(e_rad)
     print *,'The corresponding radial gridpoint', e_rad
   endif

end SUBROUTINE mp_Initialize_RHS_2d


!============================================================================


SUBROUTINE MP_MHD_PML_2d


 ! NON-CONSERVATIVE FORM OF THE FLUXES.

 implicit none
 integer i,j,k,pmlindex
 real*8 tempxy, blip
 real*8, dimension(nx,dim2(rank),nz) :: temp1, temp, temp3

 temp = - (v_x * boz - v_z * box)

 ! COMPUTE ALL z-derivatives and account for the PML
 
 call ddz(v_z, dvzdz, 1)
 call ddz(p, gradp_z, 1)

 call ddz(bx, dzbx, 1)
 call ddz(temp, dzflux2, 1)

 call PML_BOUNDARY_ACCOUNTING_2D ()

 ! ALL x-derivatives
 
 call ddx(v_x, dvxdx, 1)
 call ddx(p, gradp_x, 1)

 call ddx(temp, RHSb_z, 1)
 call ddx(bz, temp, 1)

 div = dvxdx + dvzdz


 ! CONTIUNUITY FLUX
 ! --------------------------------------
 RHScont = - gradrho0_z * v_z - rho0 * div &
	   - gradrho0_x * v_x 
           
 RHSp = - c2rho0 * div - v_x * gradp0_x &
	- v_z * gradp0_z 


 temp1 =   boz * (dzbx - temp) + bz * curlboy
 temp3 = - box * (dzbx - temp) - bx * curlboy
 
 RHSv_x = rhoinv *(temp1 - gradp_x) 
 RHSv_z = rhoinv *(temp3 - gradp_z) 

 do k=1,nz
  RHSv_z(:,:,k) = RHSv_z(:,:,k)  - rhoinv(:,:,k) * rho(:,:,k) * g(k) 

  if (abs(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) + forcing*source_dep(k)
  
 enddo

 RHSb_x =  - dzflux2

 do k=1,8
  scr(:,:,1,k) = 0.
  scr(:,:,nz,k) = 0.
 enddo

 END SUBROUTINE MP_MHD_PML_2d

!================================================================================================

SUBROUTINE CREATE_2D_SLICES ()

 implicit none
 integer i, j, k
 real*8 tempxy
 real*8, allocatable, dimension(:,:,:) :: temp

 allocate(temp(nx, nx, nz))
 call readfits_1proc(dirbackmag//'density.fits',temp,nx,nx,nz)
 rho0(:,1,:) = temp(:,nx/2,:)/dimrho
 
 call readfits_1proc(dirbackmag//'pressure.fits',temp,nx,nx,nz)
 p0(:,1,:) = temp(:,nx/2,:)/(dimrho * dimc**2.0)

 call readfits_1proc(dirbackmag//'B_x.fits',temp,nx,nx,nz)
 box(:,1,:) = temp(:,nx/2,:)/dimb

 call readfits_1proc(dirbackmag//'B_z.fits',temp,nx,nx,nz)
 boz(:,1,:) =  temp(:,nx/2,:)/dimb

 do j=1,nz
  c2(:,:,j) = gamma(j)*p0(:,:,j)/rho0(:,:,j)
 enddo
 c_speed= c2**(0.5)

 deallocate(temp)

  
END SUBROUTINE CREATE_2D_SLICES
!================================================================================================

SUBROUTINE FILTER_VARS_2D
  implicit none
  integer i,k,l,j,kst
  real*8 coeffs(0:5)
  real*8, allocatable, dimension(:,:,:,:) :: temp
  real*8, allocatable, dimension(:,:,:) :: tempfilt
  complex*16, allocatable, dimension(:,:,:) :: coeff 
  
  coeffs(0) = 0.75390625000000
  coeffs(1) = 0.41015625000000
  coeffs(2) = -0.23437500000000
  coeffs(3) = 0.08789062500000
  coeffs(4) = -0.01953125000000
  coeffs(5) = 0.00195312500000
  
  allocate(temp(nx,dim2(rank),nz,dimfive))
  kst = 6

  do l=1,dimfive
   do k=6,nz-5
    temp(:,:,k,l) = coeffs(0)*a(:,:,k,l) + 0.5*coeffs(1)*(a(:,:,k-1,l) + a(:,:,k+1,l)) &
                +  0.5*coeffs(2)*(a(:,:,k-2,l) + a(:,:,k+2,l))  &
                +  0.5*coeffs(3)*(a(:,:,k-3,l) + a(:,:,k+3,l))  &
                +  0.5*coeffs(4)*(a(:,:,k-4,l) + a(:,:,k+4,l))  &
                +  0.5*coeffs(5)*(a(:,:,k-5,l) + a(:,:,k+5,l))
   enddo

   temp(:,:,nz-4,l) = 0.5 * (a(:,:,nz-3,l) + a(:,:,nz-5,l))
   temp(:,:,nz-3,l) = 0.5 * (a(:,:,nz-2,l) + a(:,:,nz-4,l))
   temp(:,:,nz-2,l) = 0.5 * (a(:,:,nz-1,l) + a(:,:,nz-3,l))
   temp(:,:,nz-1,l) = 0.5 * (a(:,:,nz,l) + a(:,:,nz-2,l))

   if (USE_PML) then
    temp(:,:,5,l) = 0.5 * (a(:,:,6,l) + a(:,:,4,l))
    temp(:,:,4,l) = 0.5 * (a(:,:,5,l) + a(:,:,3,l))
    temp(:,:,3,l) = 0.5 * (a(:,:,4,l) + a(:,:,2,l))
    temp(:,:,2,l) = 0.5 * (a(:,:,3,l) + a(:,:,1,l))
    kst = 2
   endif

  enddo

  do l=1,dimfive
   do k=kst,nz-1
    a(:,:,k,l) = temp(:,:,k,l)
   enddo
  enddo

  deallocate(temp)

  if (( TEST_IN_2d) .and. mod(time+1,freq_filtering) .EQ. 0) then
   allocate(tempfilt(nx,dim2(rank),nz),coeff(nx/2+1,dim2(rank),nz))

   do l=1,5
    call dfftw_execute_dft_r2c(fftw_plan_fwd_x, a(:,:,:,l), coeff)

    do k=1,nz
     do j=1,dim2(rank)
      do i=1,nx/2+1
       coeff(i,j,k) = coeff(i,j,k)*decay_coeff_x(i)
       enddo
     enddo
    enddo

    call dfftw_execute_dft_c2r(fftw_plan_inv_x, coeff, a(:,:,:,l))

   enddo
  
   deallocate(tempfilt,coeff)

  endif 

END SUBROUTINE FILTER_VARS_2D

!================================================================================================


SUBROUTINE MP_MHD_SPONGE_2D

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer i,j,k
 real*8 tempxy, blip, alf(nx,dim2(rank))
 real*8 temp1(nx,dim2(rank),nz), temp2(nx,dim2(rank),nz), temp3(nx,dim2(rank),nz)
 real*8 flux1(nx,dim2(rank),nz), flux2(nx,dim2(rank),nz), flux3(nx,dim2(rank),nz)

 call ddx(v_x, dvxdx, 1)
 call ddz(v_z, dvzdz, 1)
 div = dvxdx + dvzdz

 call ddx(p, gradp_x, 1)

 do j=1,dim2(rank)
  do i=1,nx
   gradp_z(i,j,1)  =   (- c_speed(i,j,1)*rho0(i,j,1)*dvzdz(i,j,1)   &
  	    & - rho(i,j,1)*g(1))*unstretch(1)
   gradp_z(i,j,nz)  =  (c_speed(i,j,nz)*rho0(i,j,nz)*dvzdz(i,j,nz)     &
	    & - rho(i,j,nz)*g(nz))*unstretch(nz)
  enddo
 enddo

 call ddz(p, gradp_z, 4)

 ! CONTIUNUITY FLUX
 ! --------------------------------------
 RHScont = - gradrho0_z * v_z - rho0 * div &
	   - spongexyz*rho - gradrho0_x * v_x 
           

 flux3 = bz * box + boz * bx

 call ddx(bz, flux1, 1)
 call ddz(bx, flux2, 0)

 temp1 =   boz * (flux2 - flux1) + bz * curlboy
 temp3 = - box * (flux2 - flux1) - bx * curlboy

 RHSv_x = rhoinv *(temp1 - gradp_x) - spongexyz * v_x
 RHSv_z = rhoinv *(temp3 - gradp_z) - spongexyz * v_z

 do k=1,nz
  RHSv_z(:,:,k) = RHSv_z(:,:,k)  - rhoinv(:,:,k) * rho(:,:,k) * g(k) 

  if (abs(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) + forcing*source_dep(k)

 enddo

 RHSp = - c2rho0 * div - v_x * gradp0_x &
	- v_z * gradp0_z - spongexyz * p 


 flux2 = - (v_x * boz - v_z * box)

 call ddz(flux2, temp2, 1)
 RHSb_x = -temp2

 call ddx(flux2, RHSb_z, 1)

 do k=1,8
  scr(:,:,1,k) = 0.0
  scr(:,:,nz,k) = 0.0
 enddo

END SUBROUTINE MP_MHD_SPONGE_2D

!================================================================================================

SUBROUTINE MP_QUIET_PML_2D

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer k,i,j,pmlindex


 call ddz(v_z, dvzdz, 1)
 call ddz(p, gradp_z, 1)

 ! PML Accounting for dvzdz and gradp_z
 call PML_BOUNDARY_ACCOUNTING_2D ()

 call ddx(v_x, dvxdx, 1)
 call ddx(p, gradp_x, 1)

 ! Compute divergence

 div = dvxdx + dvzdz

 ! RHS continuity and pressure

 RHScont =  - rho0 * div - gradrho0_z * v_z

 RHSp = - c2rho0 * div - v_z * gradp0_z

! RHS v_x

 RHSv_x = - rhoinv * gradp_x 

! RHS v_z and PML accounting for gradp_z

 do k= 1,nz
  RHSv_z(:,:,k) =  - rhoinv(:,:,k) * (gradp_z(:,:,k) + rho(:,:,k)*g(k))

  if (ABS(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
 enddo

 RHSv_x(:,:,nz) = 0.
 RHSv_z(:,:,nz) = 0.
 RHSv_x(:,:,1) = 0.
 RHSv_z(:,:,1) = 0.
 RHSp(:,:,1) = 0.
 RHSp(:,:,nz) = 0.
 RHScont(:,:,1) = 0.
 RHScont(:,:,nz) = 0.

 END SUBROUTINE MP_QUIET_PML_2D

!================================================================================================

SUBROUTINE MP_QUIET_SPONGE_2D

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer k,i

 call ddx(v_x, dvxdx, 1)
 call ddz(v_z, dvzdz, 1)

 call ddx(p, gradp_x, 1)
 call ddz(p, gradp_z, 1)

 div = dvxdx + dvzdz
 
 RHScont = - gradrho0_z * v_z - rho0 * div - spongexyz*rho            

 RHSv_x = - rhoinv * gradp_x - spongexyz * v_x

 RHSv_z = - rhoinv * gradp_z - spongexyz * v_z

 do k= 1,nz
   RHSv_z(:,:,k) = RHSv_z(:,:,k) - rho(:,:,k)*g(k)*rhoinv(:,:,k)
   if (ABS(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
 enddo

 RHSp = - c2rho0 * div - v_z * gradp0_z  - spongexyz * p

 RHSv_x(:,:,nz) = 0.
 RHSv_z(:,:,nz) = 0.
 RHSv_x(:,:,1) = 0.
 RHSv_z(:,:,1) = 0.
 RHSp(:,:,1) = 0.
 RHSp(:,:,nz) = 0.

 END SUBROUTINE MP_QUIET_SPONGE_2D


!================================================================================================

SUBROUTINE PML_BOUNDARY_ACCOUNTING_2D
 implicit none
 integer pmlindex, k
 real*8 tempfilt(nx, dim2(rank), npmltop)
 
 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsip(:,:,pmlindex) = az(:,:,k)*gradp_z(:,:,k) + bzpml(:,:,k)*psip(:,:,pmlindex)    
   gradp_z(:,:,k) =  gradp_z(:,:,k) + psip(:,:,pmlindex)

   RHSpsivz(:,:,pmlindex) = az(:,:,k)*dvzdz(:,:,k) + bzpml(:,:,k)*psivz(:,:,pmlindex)    
   dvzdz(:,:,k) = dvzdz(:,:,k) + psivz(:,:,pmlindex)
  endif 

 enddo

 if (magnetic) then

  pmlindex = 0 
  do k=2,npmlbot
   pmlindex = pmlindex + 1
   tempfilt(:,:,pmlindex) = 0.5 * (dvzdz(:,:,k-1) + dvzdz(:,:,k+1))
  enddo

  dvzdz(:,:,2:npmlbot) = tempfilt(:,:,1:npmlbot-1)
 
  pmlindex = 0 
  do k=nz-npmltop+1,nz-1
   pmlindex = pmlindex + 1
   tempfilt(:,:,pmlindex) = 0.5 * (dvzdz(:,:,k-1) + dvzdz(:,:,k+1))
  enddo

  dvzdz(:,:,nz-npmltop+1:nz-1) = tempfilt(:,:,1:npmltop-1)
 

  pmlindex=0
  do k=1,nz

   if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
    pmlindex = pmlindex + 1
    RHSpsiinductionbx(:,:,pmlindex) = az(:,:,k)*dzflux2(:,:,k) &
			+ bzpml(:,:,k)*psiinductionbx(:,:,pmlindex)    

    dzflux2(:,:,k) = dzflux2(:,:,k) + psiinductionbx(:,:,pmlindex)

    RHSpsidzbx(:,:,pmlindex) = az(:,:,k)*dzbx(:,:,k) &
			+ bzpml(:,:,k)*psidzbx(:,:,pmlindex)    

    dzbx(:,:,k) = dzbx(:,:,k) + psidzbx(:,:,pmlindex)

   endif 
 
  enddo

 endif

END SUBROUTINE PML_BOUNDARY_ACCOUNTING_2D

!================================================================================================

END MODULE MP_PHYSICS_2D

